
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
// OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// Note: This project has gone from C++ (when it was ported from pypgdb) to C, back to C++ (where it will stay).  If
// you are making modifications, feel free to move variable declarations from the top of functions to where they are
// actually used.

#include "pyodbc.h"
#include "cursor.h"
#include "pyodbcmodule.h"
#include "connection.h"
#include "row.h"
#include "buffer.h"

#ifdef __GNUC__
typedef unsigned char byte;
typedef long long __int64;
typedef unsigned long long UINT64;
#define min(X,Y) ((X) < (Y) ? (X) : (Y))
#define max(X,Y) ((X) > (Y) ? (X) : (Y))
#define _alloca alloca
void _strlwr(char* name)
{
    while (*name) { *name = tolower(*name); name++; }
}
#endif

bool decimal_to_struct(PyObject* pTuple, SQL_NUMERIC_STRUCT& ns);

enum
{
    CURSOR_REQUIRE_CNXN    = 0x00000001,
    CURSOR_REQUIRE_OPEN    = 0x00000003, // includes _CNXN
    CURSOR_REQUIRE_RESULTS = 0x00000007, // includes _OPEN
    CURSOR_RAISE_ERROR     = 0x00000010,

    // The maximum string size we'll allocate when reading string columns if we don't know the data size.
    //
    // This is *not* the maximum string size!  It is the maximum string size we'll allocate before we know the size we
    // need.
    // 
    // When reading strings, we know the maximum size but not the data size.  If the maximum size is smaller than this
    // value, we'll allocate a string with this value, read the data, then resize the string.  If the maximum size is
    // larger than this, we'll read into a temporary buffer (_alloca(MAX_VAR_STACK)) and create a string when we know
    // the real size.
    //
    // I don't know what the right size for this is: would memory be fragmented if this was too large (how does the
    // Python allocator work with strings and resizes); what is the average string column size, etc?
    MAX_STR_ALLOCATE = 768,

    // The maximum amount of stack we use for unbound variable length columns.  We'll read in this much data to the
    // stack and then copy it to the final buffer when we know the final length.
    MAX_VAR_STACK = 2048,

    // The largest binary parameter size we'll try to bind.  If a buffer is larger than this, we'll use SQLPutData.
    MAX_VARBINARY_BUFFER = 4000,
};

struct ColumnInfo
{
    SQLSMALLINT sql_type;

    // The column size from SQLDescribeCol.  For character types, this is the maximum length, not including the NULL
    // terminator.  For binary values, this is the maximum length.  For numeric and decimal values, it is the defined
    // number of digits. For example, the precision of a column defined as NUMERIC(10,3) is 10.
    //
    // This value can be SQL_NO_TOTAL in which case the driver doesn't know the maximum length, such as for LONGVARCHAR
    // fields.
    SQLUINTEGER column_size;

    // Tells us if an integer type is signed or unsigned.  This is determined after a query using SQLColAttribute.  All
    // of the integer types are the same size whether signed and unsigned, so we can allocate memory ahead of time
    // without knowing this.  We use this during the fetch when converting to a Python integer or long.
    bool is_unsigned;
};

struct Cursor
{
    PyObject_HEAD

    // The connection object that created this cursor.
    PyObject* cnxn;

    // Set to SQL_NULL_HANDLE when the cursor is closed.
    HSTMT hstmt;

    // An array of ColumnInfos, allocated via malloc after a successful query.  This will be zero when closed or when
    // there are no query results.
    ColumnInfo* colinfos;

    // The description tuple described in the DB API 2.0 specification.  Set to None when there are no results.
    PyObject* description;

    int arraysize;

    // The Cursor.rowcount attribute from the DB API specification.
    int rowcount;

    // A dictionary that maps from column name (PyString) to index into the result columns (PyInteger).  This is
    // constructued during an execute and shared with each row (reference counted) to implement accessing results by
    // column name.
    //
    // This duplicates some ODBC functionality, but allows us to use Row objects after the statement is closed and
    // should use less memory than putting each column into the Row's __dict__.
    //
    // Since this is shared by Row objects, it cannot be reused.  New dictionaries are created for every execute.  This
    // will be zero whenever there are no results.
    PyObject* map_name_to_index;
};

inline bool
StatementIsValid(Cursor* cursor)
{
    return cursor->cnxn != 0 && ((Connection*)cursor->cnxn)->hdbc != SQL_NULL_HANDLE && cursor->hstmt != SQL_NULL_HANDLE;
}

inline Connection*
GetConnection(Cursor* cursor)
{
    return (Connection*)cursor->cnxn;
}

extern PyTypeObject CursorType;

inline bool
Cursor_Check(PyObject* o)
{
    return o != 0 && o->ob_type == &CursorType;
}


Cursor* Cursor_Validate(PyObject* obj, DWORD flags)
{
    //  Validates that a PyObject is a Cursor (like Cursor_Check) and optionally some other requirements controlled by
    //  `flags`.  If valid and all requirements (from the flags) are met, the cursor is returned, cast to Cursor*.
    //  Otherwise zero is returned.
    // 
    //  Designed to be used at the top of methods to convert the PyObject pointer and perform necessary checks.
    // 
    //  Valid flags are from the CURSOR_ enum above.  Note that unless CURSOR_RAISE_ERROR is supplied, an exception
    //  will not be set.  (When deallocating, we really don't want an exception.)

    Connection* cnxn   = 0;
    Cursor*     cursor = 0;

    if (!Cursor_Check(obj))
    {
        if (flags & CURSOR_RAISE_ERROR)
            PyErr_SetString(ProgrammingError, "Invalid cursor object.");
        return 0;
    }

    cursor = (Cursor*)obj;
    cnxn   = (Connection*)cursor->cnxn;

    if (cnxn == 0)
    {
        if (flags & CURSOR_RAISE_ERROR)
            PyErr_SetString(ProgrammingError, "Attempt to use a closed cursor.");
        return 0;
    }

    if (IsSet(flags, CURSOR_REQUIRE_OPEN))
    {
        if (cursor->hstmt == SQL_NULL_HANDLE)
        {
            if (flags & CURSOR_RAISE_ERROR)
                PyErr_SetString(ProgrammingError, "Attempt to use a closed cursor.");
            return 0;
        }

        if (cnxn->hdbc == SQL_NULL_HANDLE)
        {
            if (flags & CURSOR_RAISE_ERROR)
                PyErr_SetString(ProgrammingError, "The cursor's connection has been closed.");
            return 0;
        }
    }

    if (IsSet(flags, CURSOR_REQUIRE_RESULTS) && cursor->colinfos == 0)
    {
        if (flags & CURSOR_RAISE_ERROR)
            PyErr_SetString(ProgrammingError, "No results.  Previous SQL was not a query.");
        return 0;
    }

    return cursor;
}


PyObject*
PythonTypeFromSqlType(const SQLCHAR* name, SQLSMALLINT type)
{
    // Returns a type object ('int', 'str', etc.) for the given ODBC C type.  This is used to populate
    // Cursor.description with the type of Python object that will be returned for each column.
    //
    // name
    //   The name of the column, only used to create error messages.
    //
    // type
    //   The ODBC C type (SQL_C_CHAR, etc.) of the column.
    //
    // The returned object does not have its reference count incremented!

    PyObject* pytype = 0;

    switch (type)
    {
    case SQL_CHAR:
    case SQL_VARCHAR:
    case SQL_LONGVARCHAR:
    case SQL_GUID:
        pytype = (PyObject*)&PyString_Type;
        break;

    case SQL_DECIMAL:
    case SQL_NUMERIC:
        pytype = (PyObject*)decimal_type;

    case SQL_REAL:
    case SQL_FLOAT:
    case SQL_DOUBLE:
        pytype = (PyObject*)&PyFloat_Type;
        break;

    case SQL_SMALLINT:
    case SQL_INTEGER:
    case SQL_TINYINT:
        pytype = (PyObject*)&PyInt_Type;
        break;
        
    case SQL_TYPE_DATE:
        pytype = (PyObject*)PyDateTimeAPI->DateType;
        break;
        
    case SQL_TYPE_TIME:
        pytype = (PyObject*)PyDateTimeAPI->TimeType;
        break;
        
    case SQL_TYPE_TIMESTAMP:
        pytype = (PyObject*)PyDateTimeAPI->DateTimeType;
        break;
        
    case SQL_BIGINT:
        pytype = (PyObject*)&PyLong_Type;
        break;
        
    case SQL_BIT:
        pytype = (PyObject*)&PyBool_Type;
        break;
        
    case SQL_BINARY:
    case SQL_VARBINARY:
    case SQL_LONGVARBINARY:
        pytype = (PyObject*)&PyBuffer_Type;
        break;


    case SQL_WCHAR:
    case SQL_WVARCHAR:
    case SQL_WLONGVARCHAR:
        pytype = (PyObject*)&PyUnicode_Type;
        break;
        
    default:
        return RaiseErrorV(0, 0, "ODBC data type %d is not supported.  Cannot read column %s.", type, (const char*)name);
    }
    
    Py_INCREF(pytype);
    return pytype;
}


static bool
create_name_map(Cursor* cur, SQLSMALLINT field_count)
{
    // Called after an execute to construct the map shared by rows.

    bool success = false;
    SQLSMALLINT nDataType;
    SQLUINTEGER nColSize;
    SQLSMALLINT nullable;
    PyObject *desc = 0, *colmap = 0, *colinfo = 0, *type = 0, *index = 0, *nullable_obj=0;
    SQLCHAR name[300];
    int i;
    SQLRETURN ret;
    bool lower = lowercase();

	I(cur->hstmt != SQL_NULL_HANDLE && cur->colinfos != 0);

    // These are the values we expect after free_results.  If this function fails, we do not modify any members, so
    // they should be set to something Cursor_close can deal with.
    I(cur->description == Py_None);
    I(cur->map_name_to_index == 0);

    if (GetConnection(cur)->hdbc == SQL_NULL_HANDLE)
    {
        RaiseErrorV(0, ProgrammingError, "The cursor's connection was closed.");
        return false;
    }

    desc   = PyTuple_New((Py_ssize_t)field_count);
    colmap = PyDict_New();
    if (!desc || !colmap)
        goto done;

    for (i = 0; i < field_count; i++)
    {
        Py_BEGIN_ALLOW_THREADS
        ret = SQLDescribeCol(cur->hstmt, (USHORT)(i + 1), name, _countof(name), 0, &nDataType, &nColSize, 0, &nullable);
        Py_END_ALLOW_THREADS

        if (GetConnection(cur)->hdbc == SQL_NULL_HANDLE)
        {
            // The connection was closed by another thread in the ALLOW_THREADS block above.
            RaiseErrorV(0, ProgrammingError, "The cursor's connection was closed.");
            goto done;
        }

        if (!SQL_SUCCEEDED(ret))
        {
            RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
            goto done;
        }

        if (lower)
            _strlwr((char*)name);

        type = PythonTypeFromSqlType(name, nDataType);
        if (!type)
            goto done;

        switch (nullable)
        {
        case SQL_NO_NULLS:
            nullable_obj = Py_False;
            break;
        case SQL_NULLABLE:
            nullable_obj = Py_True;
            break;
        case SQL_NULLABLE_UNKNOWN:
        default:
            nullable_obj = Py_None;
            break;
        }

        Py_INCREF(nullable_obj);

        colinfo = Py_BuildValue("(sOsissO)", (char*)name, type, 0, (int)nColSize, 0, 0, nullable_obj);
        if (!colinfo)
            goto done;

        nullable_obj = 0;

        index = PyInt_FromLong(i);
        if (!index)
            goto done;

        PyDict_SetItemString(colmap, (const char*)name, index);
        Py_DECREF(index);       // SetItemString increments
        index = 0;

        PyTuple_SET_ITEM(desc, i, colinfo);
        colinfo = 0;            // reference stolen by SET_ITEM
    }

    Py_XDECREF(cur->description);
    cur->description = desc;
    desc = 0;
    cur->map_name_to_index = colmap;
    colmap = 0;

    success = true;

  done:
    Py_XDECREF(nullable_obj);
    Py_XDECREF(desc);
    Py_XDECREF(colmap);
    Py_XDECREF(index);
    Py_XDECREF(colinfo);

    return success;
}


static bool
free_results(Cursor* self)
{
    // Internal function called anytime we need to free the memory associated with query results.  It is safe to call
    // this even when a query has not been executed.

    // If we ran out of memory, it is possible that we have a cursor but colinfos is zero.  However, we should be
    // deleting this object, so the cursor will be freed when the HSTMT is destroyed. */

    SQLRETURN ret;

    if (self->colinfos)
    {
        if (StatementIsValid(self))
        {
            Py_BEGIN_ALLOW_THREADS
            ret = SQLCloseCursor(self->hstmt);
            Py_END_ALLOW_THREADS
            
            if (GetConnection(self)->hdbc == SQL_NULL_HANDLE)
            {
                // The connection was closed by another thread in the ALLOW_THREADS block above.
                RaiseErrorV(0, ProgrammingError, "The cursor's connection was closed.");
                return false;
            }
        }
        
        free(self->colinfos);
        self->colinfos = 0;
    }

    if (self->description != Py_None)
    {
        Py_DECREF(self->description);
        self->description = Py_None;
        Py_INCREF(Py_None);
    }

    if (self->map_name_to_index)
    {
        Py_DECREF(self->map_name_to_index);
        self->map_name_to_index = 0;
    }

    self->rowcount = -1;

    return true;
}


static void
closeimpl(Cursor* cur)
{
    // An internal function for the shared 'closing' code used by Cursor_close and Cursor_dealloc.
    // 
    // This method releases the GIL lock while closing, so verify the HDBC still exists if you use it.

    free_results(cur);

    if (StatementIsValid(cur))
    {
        HSTMT hstmt = cur->hstmt;
        cur->hstmt = SQL_NULL_HANDLE;
        Py_BEGIN_ALLOW_THREADS
        SQLFreeHandle(SQL_HANDLE_STMT, hstmt);
        Py_END_ALLOW_THREADS
    }

    Py_XDECREF(cur->description);
    Py_XDECREF(cur->map_name_to_index);
    Py_XDECREF(cur->cnxn);

    cur->description = 0;
    cur->map_name_to_index = 0;
    cur->cnxn = 0;
}

static char close_doc[] =
    "Close the cursor now (rather than whenever __del__ is called).  The cursor will\n" 
    "be unusable from this point forward; a ProgrammingError exception will be\n" 
    "raised if any operation is attempted with the cursor.";
    
static PyObject*
Cursor_close(PyObject* self, PyObject* args)
{
    Cursor* cursor = Cursor_Validate(self, CURSOR_REQUIRE_OPEN | CURSOR_RAISE_ERROR);
    if (!cursor)
        return 0;

    closeimpl(cursor);

    Py_INCREF(Py_None);
    return Py_None;
}

static void
Cursor_dealloc(Cursor* cursor)
{
    if (Cursor_Validate((PyObject*)cursor, CURSOR_REQUIRE_CNXN))
    {
        closeimpl(cursor);
    }

    PyObject_Del(cursor);
}


static int
GetParamBufferSize(PyObject* param, Py_ssize_t iParam)
{
    // Returns the size in bytes needed to hold the parameter in a format for binding, used to allocate the parameter
    // buffer.  (The value is not passed to ODBC.  Values passed to ODBC are in BindParam.)
    //
    // If we can bind directly into the Python object (e.g., using PyString_AsString), zero is returned since no extra
    // memory is required.  If the data will be provided at execution time (e.g. SQL_DATA_AT_EXEC), zero is returned
    // since the parameter value is not stored at all.  If the data type is not recognized, -1 is returned.

    if (param == Py_None)
        return 0;

    if (PyString_Check(param) || PyUnicode_Check(param))
        return 0;

    if (param == Py_True || param == Py_False)
        return 1;

    if (PyDateTime_Check(param))
        return sizeof(TIMESTAMP_STRUCT);

    if (PyDate_Check(param))
        return sizeof(DATE_STRUCT);

    if (PyTime_Check(param))
        return sizeof(TIME_STRUCT);

    if (PyInt_Check(param))
        return sizeof(long int);

    if (PyLong_Check(param))
        return sizeof(__int64);

    if (PyFloat_Check(param))
        return sizeof(double);

    if (PyDecimal_Check(param))
        return sizeof(SQL_NUMERIC_STRUCT);

    if (PyBuffer_Check(param))
    {
        // If the buffer has a single segment, we can bind directly to it, so we need 0 bytes.  Otherwise, we'll use
        // SQL_DATA_AT_EXEC, so we still need 0 bytes.
        return 0;
    }

    RaiseErrorV("HY105", ProgrammingError, "Invalid parameter type.  param-index=%zd param-type=%s", iParam, param->ob_type->tp_name);

    return -1;
}


bool
InitColumnInfo(Cursor* cursor, SQLUSMALLINT iCol, ColumnInfo* pinfo)
{
    // Initializes ColumnInfo from result set metadata.

    int f;
    SQLRETURN ret;

    Py_BEGIN_ALLOW_THREADS
    ret = SQLDescribeCol(cursor->hstmt, iCol,
                         0, 0, 0,
                         &pinfo->sql_type,
                         &pinfo->column_size,
                         0,
                         0);
    Py_END_ALLOW_THREADS

    if (GetConnection(cursor)->hdbc == SQL_NULL_HANDLE)
    {
        // The connection was closed by another thread in the ALLOW_THREADS block above.
        RaiseErrorV(0, ProgrammingError, "The cursor's connection was closed.");
        return false;
    }

    if (!SQL_SUCCEEDED(ret))
    {
        RaiseErrorFromHandle(GetConnection(cursor)->hdbc, cursor->hstmt);
        return false;
    }

    // If it is an integer type, determine if it is signed or unsigned.  The buffer size is the same but we'll need to
    // know when we convert to a Python integer.

    switch (pinfo->sql_type)
    {
    case SQL_TINYINT:
    case SQL_SMALLINT:
    case SQL_INTEGER:
    case SQL_BIGINT:
    {
        Py_BEGIN_ALLOW_THREADS
        ret = SQLColAttribute(cursor->hstmt, iCol, SQL_DESC_UNSIGNED, 0, 0, 0, &f);
        Py_END_ALLOW_THREADS

        if (GetConnection(cursor)->hdbc == SQL_NULL_HANDLE)
        {
            // The connection was closed by another thread in the ALLOW_THREADS block above.
            RaiseErrorV(0, ProgrammingError, "The cursor's connection was closed.");
            return false;
        }

        if (!SQL_SUCCEEDED(ret))
        {
            RaiseErrorFromHandle(GetConnection(cursor)->hdbc, cursor->hstmt);
            return false;
        }
        pinfo->is_unsigned = (f == SQL_TRUE);
        break;
    }
    
    default:
        pinfo->is_unsigned = false;
    }

    return true;
}


static bool
PrepareResults(Cursor* cur, int cCols)
{
    // Called after a SELECT has been executed to perform pre-fetch work.
    // 
    // Allocates the ColumnInfo structures describing the returned data.

    int i;
    I(cur->colinfos == 0);

    cur->colinfos = (ColumnInfo*)malloc(sizeof(ColumnInfo) * cCols);
    if (cur->colinfos == 0)
    {
        PyErr_NoMemory();
        return false;
    }

    for (i = 0; i < cCols; i++)
    {
        if (!InitColumnInfo(cur, (SQLSMALLINT)(i + 1), &cur->colinfos[i]))
        {
            free(cur->colinfos);
            cur->colinfos = 0;
            return false;
        }
    }

    return true;
}

static bool
BindParam(Cursor* cur, int iParam, PyObject* param, int* pibNext, byte* pbParams)
{
    // When binding, ODBC requires 2 values: the column size and the buffer size.  The column size is related to the
    // destination (the SQL type of the column being written to) and the buffer size refers to the source (the size of
    // the C data being written).  If you send the wrong column size, data may be truncated.  For example, if you send
    // sizeof(TIMESTAMP_STRUCT) as the column size when writing a timestamp, it will be rounded to the nearest minute
    // because that is the precision that would fit into a string of that size.

    // This function is a bit finicky, so be careful.  It must match the loop that allocated the memory pibNext points
    // into.  That means that pibNext should not be incremented if we are binding directly into a Python object
    // (e.g. strings) since we did not allocate any memory.

    SQLRETURN ret;
    SQLSMALLINT ctype     = 0;
    SQLSMALLINT sqltype   = 0;
    SQLUINTEGER colsize   = 0;
    SQLSMALLINT decdigits = 0;
    SQLPOINTER  pbValue   = 0;
    SQLINTEGER  bufsize   = 0;

    const char* pb;
    Py_ssize_t cb;

    // If you don't want to use pcbData, set it to zero.  (Not its value -- the pointer.)
    SQLINTEGER* pcbData = (SQLINTEGER*)&pbParams[*pibNext];
    *pibNext += sizeof(SQLINTEGER);

    if (param == Py_None)
    {
        // Note: We don't actually know what the destination data type is (or the source for that matter), so we choose
        // character since every type has a character representation.  (That is, it is always valid to write a
        // character string to a field.)

        // The column size has to be larger than the greatest minimum size for all of the different data types.  For
        // example, when writing a NULL to a datetime column, the colsize will be compared to the actual destination
        // type (timestamp) and must be at least 16.  I've only tested against SQL Server, but it doesn't seem to mind
        // that the size is a too large.  (SQL Server only supports colsize up to 23 for datetime unless writing a
        // NULL.)  The value 36 below is the largest in Appendix D of the ODBC manual and is the length of a GUID in
        // string form.

        ctype   = SQL_C_CHAR;
        sqltype = SQL_VARCHAR;
        pbValue = 0;
        colsize = 36;
        bufsize = 0;
        *pcbData = SQL_NULL_DATA;
    }
    else if (PyString_Check(param))
    {
        ctype    = SQL_C_CHAR;
        pbValue  = PyString_AS_STRING(param);
        *pcbData = (SQLINTEGER)PyString_GET_SIZE(param);
        colsize  = max(1, *pcbData); // Note: We cannot set colsize to zero ([HY104] Invalid precision value (0))
        bufsize  = colsize;
        sqltype  = (colsize <= 255) ? SQL_VARCHAR : SQL_LONGVARCHAR;
    }
    else if (PyUnicode_Check(param))
    {
        ctype    = SQL_C_WCHAR;
        pbValue  = PyUnicode_AsUnicode(param);
        *pcbData = (SQLINTEGER)PyUnicode_GET_SIZE(param) * sizeof(wchar_t);
        colsize  = max(sizeof(wchar_t), (unsigned int)*pcbData); // Note: We cannot set colsize to zero ([HY104] Invalid precision value (0))
        bufsize  = bufsize;
        sqltype  = (colsize <= 255) ? SQL_WVARCHAR : SQL_WLONGVARCHAR;
    }
    else if (param == Py_True || param == Py_False)
    {
        pbParams[*pibNext] = (unsigned char)(param == Py_True ? 1 : 0);

        ctype   = SQL_C_BIT;
        sqltype = SQL_BIT;
        pbValue = &pbParams[*pibNext++];
        bufsize = 1;
        pcbData = 0;

        *pibNext += bufsize;
    }
    else if (PyDateTime_Check(param))
    {
        TIMESTAMP_STRUCT* value = (TIMESTAMP_STRUCT*)&pbParams[*pibNext];

        value->year     = (SQLSMALLINT) PyDateTime_GET_YEAR(param);
        value->month    = (SQLUSMALLINT)PyDateTime_GET_MONTH(param);
        value->day      = (SQLUSMALLINT)PyDateTime_GET_DAY(param);
        value->hour     = (SQLUSMALLINT)PyDateTime_DATE_GET_HOUR(param);
        value->minute   = (SQLUSMALLINT)PyDateTime_DATE_GET_MINUTE(param);
        value->second   = (SQLUSMALLINT)PyDateTime_DATE_GET_SECOND(param);
        value->fraction = (SQLUINTEGER)(PyDateTime_DATE_GET_MICROSECOND(param) * 1000); // 1000 == micro -> nano

        // When I accidentally set colsize to 24, an exception was thrown indicating it was out of range.  Since the
        // documentation doesn't specify a range, it appears this might be set by the driver.  How can we find out what
        // is supported so we can actually use different drivers?

        ctype   = SQL_C_TIMESTAMP;
        sqltype = SQL_TYPE_TIMESTAMP;
        pbValue = value;
        bufsize = sizeof(TIMESTAMP_STRUCT);
        colsize = 23;           // .001 second precision, which is what SQL Server supports
        pcbData = 0;

        *pibNext += bufsize;
    }
    else if (PyDate_Check(param))
    {
        DATE_STRUCT* value = (DATE_STRUCT*)&pbParams[*pibNext];
        value->year  = (SQLSMALLINT) PyDateTime_GET_YEAR(param);
        value->month = (SQLUSMALLINT)PyDateTime_GET_MONTH(param);
        value->day   = (SQLUSMALLINT)PyDateTime_GET_DAY(param);

        ctype   = SQL_C_DATE;
        sqltype = SQL_TYPE_DATE;
        pbValue = value;
        bufsize = sizeof(DATE_STRUCT);
        pcbData = 0;

        *pibNext += bufsize;
    }
    else if (PyTime_Check(param))
    {
        TIME_STRUCT* value = (TIME_STRUCT*)&pbParams[*pibNext];
        value->hour     = (SQLUSMALLINT)PyDateTime_TIME_GET_HOUR(param);
        value->minute   = (SQLUSMALLINT)PyDateTime_TIME_GET_MINUTE(param);
        value->second   = (SQLUSMALLINT)PyDateTime_TIME_GET_SECOND(param);

        ctype   = SQL_C_TIME;
        sqltype = SQL_TYPE_TIME;
        pbValue = value;
        bufsize = sizeof(TIME_STRUCT);
        pcbData = 0;

        *pibNext += bufsize;
    }
    else if (PyInt_Check(param))
    {
        long* value = (long*)&pbParams[*pibNext];

        *value = PyInt_AsLong(param);

        ctype   = SQL_C_LONG;
        sqltype = SQL_INTEGER;
        pbValue = value;
        bufsize = sizeof(long);
        pcbData = 0;

        *pibNext += bufsize;
    }
    else if (PyLong_Check(param))
    {
        __int64* value = (__int64*)&pbParams[*pibNext];

        *value = PyInt_AsLong(param);

        ctype   = SQL_C_SBIGINT;
        sqltype = SQL_BIGINT;
        pbValue = value;
        bufsize = sizeof(__int64);
        pcbData = 0;

        *pibNext += bufsize;
    }
    else if (PyFloat_Check(param))
    {
        double* value = (double*)&pbParams[*pibNext];

        *value = PyFloat_AsDouble(param);

        ctype   = SQL_C_DOUBLE;
        sqltype = SQL_DOUBLE;
        pbValue = value;
        bufsize = sizeof(double);
        pcbData = 0;

        *pibNext += bufsize;
    }
    else if (PyDecimal_Check(param))
    {
        PyObject* pTuple = PyObject_CallMethod(param, "as_tuple", 0);
        if (!pTuple)
            return false;

        SQL_NUMERIC_STRUCT& ns = *(SQL_NUMERIC_STRUCT*)&pbParams[*pibNext];
        
        if (!decimal_to_struct(pTuple, ns))
        {
            Py_DECREF(pTuple);
            return false;
        }

        Py_DECREF(pTuple);

        ctype     = SQL_C_NUMERIC;
        sqltype   = SQL_DECIMAL;
        pbValue   = &ns;
        bufsize   = sizeof(SQL_NUMERIC_STRUCT);
        colsize   = ns.precision;
        decdigits = ns.scale;
        pcbData   = 0;

        *pibNext += bufsize;
    }
    else if (PyBuffer_Check(param))
    {
        cb = PyBuffer_GetMemory(param, &pb);

        if (cb != -1 && cb <= MAX_VARBINARY_BUFFER)
        {
            // There is one segment, so we can bind directly into the buffer object.

            ctype   = SQL_C_BINARY;
            sqltype = SQL_VARBINARY;
    
            pbValue  = (SQLPOINTER)pb;
            bufsize  = cb;
            colsize  = cb;
            *pcbData = cb;
        }
        else
        {
            // There are multiple segments, so we'll provide the data at execution time.  Pass the PyObject pointer as
            // the parameter value which will be pased back to us when the data is needed.  (If we release threads, we
            // need to up the refcount!)

            ctype   = SQL_C_BINARY;
            sqltype = SQL_LONGVARBINARY;

            pbValue  = param;
            colsize  = PyBuffer_Size(param);
            bufsize  = sizeof(PyObject*); // How big is pbValue; ODBC copies it and gives it back in SQLParamData
            *pcbData = SQL_LEN_DATA_AT_EXEC(PyBuffer_Size(param));
        }
    }
    else
    {
        RaiseErrorV("HY097", NotSupportedError, "Python type %s not supported.  param=%d", param->ob_type->tp_name, iParam);
        return false;
    }

    Py_BEGIN_ALLOW_THREADS
    ret = SQLBindParameter(cur->hstmt, (USHORT)iParam, SQL_PARAM_INPUT, ctype, sqltype, colsize, decdigits, pbValue, bufsize, pcbData);

    if (ctype == SQL_C_NUMERIC && SQL_SUCCEEDED(ret) && GetConnection(cur)->odbc_major >= 3)
    {
        // The ODBC documentation seems to say that the precision and scale in the SQL_NUMERIC_STRUCT are ignored for
        // input and we must execute the following code to set these values.  At least some versions of the Oracle
        // driver use the values from the structure instead, so we set both.

        // I've put the odbc_major check here (which is why odbc_major was added; remove it if you remove this check)
        // because some older drivers don't event like the GetStmtAttr call.

        SQLHDESC hdesc;
        ret = SQLGetStmtAttr(cur->hstmt, SQL_ATTR_APP_ROW_DESC, &hdesc, 0, NULL);
        if (SQL_SUCCEEDED(ret))
        {
            // Not all drivers like this (PostgreSQL Linux for example), so I'm going to ignore it if it fails.
        ret = SQLSetDescField(hdesc, 1, SQL_DESC_TYPE, (void*)SQL_C_NUMERIC, 0);
        ret = SQLSetDescField(hdesc, 1, SQL_DESC_PRECISION, (void*)colsize, 0);
        ret = SQLSetDescField(hdesc, 1, SQL_DESC_SCALE, (void*) decdigits, 0);        
        }
        else
        {
            ret = SQL_SUCCESS;
        }
    }
    
    Py_END_ALLOW_THREADS

    if (GetConnection(cur)->hdbc == SQL_NULL_HANDLE)
    {
        // The connection was closed by another thread in the ALLOW_THREADS block above.
        RaiseErrorV(0, ProgrammingError, "The cursor's connection was closed.");
        return false;
    }

    if (!SQL_SUCCEEDED(ret))
    {
        RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
        return false;
    }

    return true;
}


static Py_ssize_t
count_params(PyObject* pSql)
{
    // Returns the number of '?' parameters in the SQL string.

    // Question marks are valid in quotes, so don't count anything in quotes.

    Py_ssize_t count = 0;
    bool in_quotes = false;

    if (PyString_Check(pSql))
    {
        const char* pch = PyString_AS_STRING(pSql);

        while (*pch)
        {
            if (*pch == '\'')
            {
                in_quotes = !in_quotes;
            }
            else if (*pch == '?' && !in_quotes)
            {
                count++;
            }

            pch++;
        }
    }
    else
    {
        // const wchar_t* pch = PyUnicode_AsUnicode(pSql);
        const Py_UNICODE* pch = PyUnicode_AsUnicode(pSql);

        while (*pch)
        {
            if (*pch == '\'')
            {
                in_quotes = !in_quotes;
            }
            else if (*pch == L'?' && !in_quotes)
            {
                count++;
            }

            pch++;
        }
    }
    
    return count;
}


static PyObject*
execute(Cursor* cur, PyObject* pSql, PyObject* params, bool skip_first)
{
    // Internal function to execute SQL, called by .execute and .executemany.
    // 
    // pSql
    //   A PyString, PyUnicode, or derived object containing the SQL.
    // 
    // skip_first
    //   If true, the first element in `params` is ignored.  (It will be the SQL statement and `params` will be the
    //   entire tuple passed to Cursor.execute.)  Otherwise all of the params are used.  (This case occurs when called
    //   from Cursor.executemany, in which case the sequences do not contain the SQL statement.)  Ignored if params is
    //   zero.

    SQLRETURN ret = 0;

    int params_offset = skip_first ? 1 : 0;
    Py_ssize_t cParams = params == 0 ? 0 : PySequence_Length(params) - params_offset;
    Py_ssize_t cb = 0;

    // If parameters are provided and we need to convert them to a different format for ODBC, memory will be allocated
    // here (using malloc).
    byte* param_buffer = 0;

    free_results(cur);

    if (cParams > 0)
    {
        // Calculate the amount of memory we need for param_buffer.  We can't allow it to reallocate on the fly since
        // we will bind directly into its memory.  (We only use a vector so its destructor will free the memory.)
        // We'll set aside one SQLINTEGER for each column to be used as the StrLen_or_IndPtr.

        PyObject* param = 0;
        int cbT = 0, ibNext = 0;  // index in param_buffer where next parameter can be placed.

        for (Py_ssize_t i = 0; i < cParams; i++)
        {
            param = PySequence_GetItem(params, i + params_offset);
            if (param == 0)
                return 0;

            cbT = GetParamBufferSize(param, i + 1) + sizeof(SQLINTEGER); // +1 to map to ODBC one-based index

            Py_DECREF(param);

            if (cbT < 0)
                return 0;

            cb += cbT;
        }

        param_buffer = (byte*)malloc(cb);
        if (!param_buffer)
            return PyErr_NoMemory();

        // Bind each parameter.  If possible, items will be bound directly into the Python object.  Otherwise,
        // param_buffer will be used and ibNext will be updated.

        for (Py_ssize_t i = 0; i < cParams; i++)
        {
            param = PySequence_GetItem(params, i + params_offset);

            if (!BindParam(cur, i + 1, param, &ibNext, &param_buffer[0]))
            {
                Py_DECREF(param);
                free(param_buffer);
                return 0;
            }

            Py_DECREF(param);
        }
    }

    if (PyString_Check(pSql))
    {
#ifdef TRACE_ALL
        printf("cursor.execute cnxn=%p hdbc=%d cursor=%p sql=%s\n", (Connection*)cur->cnxn, ((Connection*)cur->cnxn)->hdbc, cur, PyString_AS_STRING(pSql));
#endif
        Py_BEGIN_ALLOW_THREADS
        ret = SQLExecDirect(cur->hstmt, (SQLCHAR*)PyString_AS_STRING(pSql), SQL_NTS);
        Py_END_ALLOW_THREADS
    }
    else
    {
        Py_BEGIN_ALLOW_THREADS
        ret = SQLExecDirectW(cur->hstmt, (SQLWCHAR*)PyUnicode_AsUnicode(pSql), SQL_NTS);
        Py_END_ALLOW_THREADS
    }
    
    free(param_buffer);

    if (GetConnection(cur)->hdbc == SQL_NULL_HANDLE)
    {
        // The connection was closed by another thread in the ALLOW_THREADS block above.
        return RaiseErrorV(0, ProgrammingError, "The cursor's connection was closed.");
    }

    while (ret == SQL_NEED_DATA)
    {
        // We have bound a `buffer` object using SQL_DATA_AT_EXEC, so ODBC is asking us for the data now.  We gave the
        // buffer pointer to ODBC in SQLBindParameter -- SQLParamData below gives the pointer back to us.

        PyObject* pBuffer;
        ret = SQLParamData(cur->hstmt, (SQLPOINTER*)&pBuffer);

        if (ret == SQL_NEED_DATA)
        {
            // Buffers can have multiple segments, so we might need multiple writes.  Looping through buffers isn't
            // difficult, but we've wrapped it up in an iterator object to keep this loop simple.

            BufferSegmentIterator it(pBuffer);
            byte* pb;
            Py_ssize_t cb;
            while (it.Next(pb, cb))
                SQLPutData(cur->hstmt, pb, cb);
        }
    }

    if (ret == SQL_NO_DATA)
    {
        // Example: A delete statement that did not delete anything.
        cur->rowcount = 0;
        return PyInt_FromLong(cur->rowcount);
    }

    if (!SQL_SUCCEEDED(ret))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    SQLSMALLINT cCols;
    if (!SQL_SUCCEEDED(SQLNumResultCols(cur->hstmt, &cCols)))
    {
        // Note: The SQL Server driver sometimes returns HY007 here if multiple statements (separated by ;) were
        // submitted.  This is not documented, but I've seen it with multiple successful inserts.  

        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
    }
    
    SQLINTEGER cRows;
    Py_BEGIN_ALLOW_THREADS
    ret = SQLRowCount(cur->hstmt, &cRows);
    Py_END_ALLOW_THREADS
    cur->rowcount = (int)cRows;

    if (cCols != 0)
    {
        // A result set was created.

        if (!PrepareResults(cur, cCols))
            return 0;

        if (!create_name_map(cur, cCols))
            return 0;

        // Return the cursor so the results can be iterated over directly.
        Py_INCREF(cur);
        return (PyObject*)cur;
    }

    if (GetConnection(cur)->hdbc == SQL_NULL_HANDLE)
    {
        // The connection was closed by another thread in the ALLOW_THREADS block above.
        return RaiseErrorV(0, ProgrammingError, "The cursor's connection was closed.");
    }

    if (!SQL_SUCCEEDED(ret))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
    
    return PyInt_FromLong(cur->rowcount);
}

inline bool
IsSequence(PyObject* p)
{
    return PySequence_Check(p) && !PyString_Check(p) && !PyBuffer_Check(p) && !PyUnicode_Check(p);
}

static char execute_doc[] =
    "C.execute(sql, [params]) --> None | Cursor | count\n"
    "\n"
    "Prepare and execute a database query or command.\n"
    "\n"
    "Parameters may be provided as a sequence (as specified by the DB API) or\n"
    "simply passed in one after another (non-standard):\n"
    "\n"
    "  cursor.execute(sql, (param1, param2))\n"
    "\n"
    "    or\n"
    "\n"
    "  cursor.execute(sql, param1, param2)\n"
    "\n"
    "The return value for this method is not specified in the API, so any use is\n"
    "non-standard.  For convenience, the type depends on the operation performed.\n"
    "A select statement will return `self` so the results can be iterated\n"
    "conveniently:\n"
    "\n"
    "  for row in cursor.execute('select * from tmp'):\n"
    "      print row.customer_id\n"
    "\n"
    "An update or delete statement will return the number of records affected as an\n"
    "integer:\n"
    "\n"
    "  count = cursor.execute('delete from tmp')\n"
    "\n"
    "If any other statement will return None.";

PyObject*
Cursor_execute(PyObject* self, PyObject* args)
{
    Py_ssize_t cParams = PyTuple_Size(args) - 1;

    bool skip_first = false;
    PyObject *pSql, *params = 0, *result = 0;

    Cursor* cursor = Cursor_Validate(self, CURSOR_REQUIRE_OPEN | CURSOR_RAISE_ERROR);
    if (!cursor)
        return 0;

#ifdef _DEBUG
    DebugBreak();
#endif
    
    if (cParams < 0)
    {
        PyErr_SetString(PyExc_TypeError, "execute() takes at least 1 argument (0 given)");
        goto done;
    }

    pSql = PyTuple_GET_ITEM(args, 0);

    if (!PyString_Check(pSql) && !PyUnicode_Check(pSql))
    {
        PyErr_SetString(PyExc_TypeError, "The first argument to execute must be a string or unicode query.");
        goto done;
    }

    // Figure out if there were parameters and how they were passed.  Our optional parameter passing complicates this slightly.

    if (cParams == 1 && IsSequence(PyTuple_GET_ITEM(args, 1)))
    {
        // There is a single argument and it is a sequence, so we must treat it as a sequence of parameters.  (This is
        // the normal Cursor.execute behavior.)

        params     = PyTuple_GET_ITEM(args, 1);
        skip_first = false;
    }
    else if (cParams > 0)
    {
        params     = args;
        skip_first = true;
    }

    // Execute.

    result = execute(cursor, pSql, params, skip_first);

  done:

    return result;
}

static PyObject*
Cursor_executemany(PyObject* self, PyObject* args)
{
    PyObject *pSql, *param_seq, *params;

    Cursor* cursor = Cursor_Validate(self, CURSOR_REQUIRE_OPEN | CURSOR_RAISE_ERROR);
    if (!cursor)
        return 0;

    cursor->rowcount = -1;

    if (!PyArg_ParseTuple(args, "OO", &pSql, &param_seq))
        return 0;

    if (!PyString_Check(pSql) && !PyUnicode_Check(pSql))
    {
        PyErr_SetString(PyExc_TypeError, "The first argument to execute must be a string or unicode query.");
        return 0;
    }

    if (!IsSequence(param_seq))
    {
        PyErr_SetString(ProgrammingError, "The second parameter to executemany must be a sequence.");
        return 0;
    }

    Py_ssize_t c = PySequence_Size(param_seq);

    if (c == 0)
    {
        PyErr_SetString(ProgrammingError, "The second parameter to executemany must not be empty.");
        return 0;
    }

    Py_ssize_t param_count = count_params(pSql);

    for (Py_ssize_t i = 0; i < c; i++)
    {
        params = PySequence_GetItem(param_seq, i);

        if (!IsSequence(params))
        {
            RaiseErrorV(0, ProgrammingError, "Parameter sequence %zd is not a sequence.  type=%s", i, params->ob_type->tp_name);
            Py_DECREF(params);
            return 0;
        }

        if (PySequence_Size(params) != param_count)
        {
            RaiseErrorV(0, ProgrammingError,
                        "Parameter sequence %zd does not have the correct number of parameters.  expected=%zd  received=%zd",
                        param_count, PySequence_Size(params));
            Py_DECREF(params);
            return 0;
        }

        Py_DECREF(params);
    }

    for (Py_ssize_t i = 0; i < c; i++)
    {
        params = PySequence_GetItem(param_seq, i);
        PyObject* result = execute(cursor, pSql, params, false);
        Py_XDECREF(result);
        Py_DECREF(params);
        if (!result)
            break;
    }

    cursor->rowcount = -1;
    Py_RETURN_NONE;
}


static PyObject*
GetDataString(Cursor* cur, int iCol)
{
    // Returns a String or Unicode object for character and binary data.

    // NULL terminator notes:
    //
    //  * pinfo->column_size, from SQLDescribeCol, does not include a NULL terminator.  For example, column_size for a
    //    char(10) column would be 10.  (Also, when dealing with wchar_t, it is the number of *characters*, not bytes.)
    //
    //  * When passing a length to PyString_FromStringAndSize and similar Unicode functions, do not add the NULL
    //    terminator -- it will be added automatically.  See objects/stringobject.c
    //
    //  * SQLGetData does not return the NULL terminator in the length indicator.  (Therefore, You can pass this value
    //    directly to the string functions.)
    //
    //  * SQLGetData will write a NULL terminator in the output buffer, so you must leave room for it.  You must also
    //    include the NULL terminator in the buffer length passed to SQLGetData.
    //
    // ODBC generalization:
    //  1) Include NULL terminators in input buffer lengths.
    //  2) NULL terminators are not used in data lengths.

    ColumnInfo* pinfo = &cur->colinfos[iCol];
    SQLSMALLINT nTargetType;
    PyObject* result = 0;
    SQLINTEGER cbData = 0;
    SQLRETURN ret;
    char* pch;
    int element_size, null_size = 0;

    // Some Unix ODBC drivers do not return the correct length.
    if (pinfo->sql_type == SQL_GUID)
        pinfo->column_size = 36;

    switch (pinfo->sql_type)
    {
    case SQL_CHAR:
    case SQL_VARCHAR:
    case SQL_LONGVARCHAR:
    case SQL_GUID:
        nTargetType = SQL_C_CHAR;
        element_size = sizeof(char);
        null_size    = sizeof(char);
        break;

    case SQL_WCHAR:
    case SQL_WVARCHAR:
    case SQL_WLONGVARCHAR:
        nTargetType = SQL_C_WCHAR;
        element_size = sizeof(wchar_t);
        null_size    = sizeof(wchar_t);
        break;
        
    default:
        nTargetType = SQL_C_BINARY;
        element_size = sizeof(char);
        null_size    = 0;
        break;
    }

    if (pinfo->column_size <= MAX_STR_ALLOCATE && (pinfo->sql_type == SQL_CHAR || pinfo->sql_type == SQL_BINARY || pinfo->sql_type == SQL_WCHAR || pinfo->sql_type == SQL_GUID))
    {
        // The result is a PyString and not too large, so we'll allocate a PyString, read the data into it, and resize
        // the string when we're finished.

        if (nTargetType == SQL_C_WCHAR)
        {
            result = PyUnicode_FromUnicode(0, pinfo->column_size);
            pch    = result ? (char*)PyUnicode_AsUnicode(result) : 0;
        }
        else
        {
            result = PyString_FromStringAndSize(0, pinfo->column_size);
            pch    = result ? PyString_AS_STRING(result) : 0;
        }

        if (!result)
            return 0;
    
        Py_BEGIN_ALLOW_THREADS
        ret = SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), nTargetType, pch,
                         pinfo->column_size * element_size + null_size, &cbData);
        Py_END_ALLOW_THREADS

        if (GetConnection(cur)->hdbc == SQL_NULL_HANDLE)
            return RaiseErrorV(0, ProgrammingError, "The cursor's connection was closed.");

        if (cbData == SQL_NULL_DATA)
        {
            Py_DECREF(result);
            Py_RETURN_NONE;
        }

        if (!SQL_SUCCEEDED(ret))
        {
            Py_DECREF(result);
            return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
        }

        // Now that we know the actual size, resize the string.  Both of these functions will deallocate result on
        // failure and set it to 0.
        if (nTargetType == SQL_C_WCHAR)
            PyUnicode_Resize(&result, cbData / element_size);
        else
            _PyString_Resize(&result, cbData);

        return result;
    }

    // We don't know the exact length, so we'll try reading into a temporary buffer.  (This is a performance
    // optimization since some drivers consider any string over 255 "long".  We expect a lot of strings to be read in a
    // single SQLGetData call.)

    char ab[MAX_VAR_STACK];

    cbData = 0;
    ret = SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), nTargetType, ab, MAX_VAR_STACK, &cbData);
    if (ret == SQL_SUCCESS)
    {
        // We read it all, so we now know the size.
    
        if (cbData == SQL_NULL_DATA)
            Py_RETURN_NONE;

        if (nTargetType == SQL_C_WCHAR)
            return PyUnicode_FromUnicode((const Py_UNICODE*)ab, cbData / element_size);
        else
            return PyString_FromStringAndSize(ab, cbData);
    }

    if (cbData != SQL_NO_TOTAL)
    {
        // The data didn't fit, but the driver returned the length in cbData.  (This includes the data already
        // copied into ab.)  Now that we know the size, we can allocate a string, copy what we have from ab to the new
        // string, and read the rest into the remaining part of the string.

        if (nTargetType == SQL_C_WCHAR)
        {
            result = PyUnicode_FromUnicode(0, cbData / element_size);
            pch    = result ? (char*)PyUnicode_AsUnicode(result) : 0;
        }
        else
        {
            result = PyString_FromStringAndSize(0, cbData);
            pch    = result ? PyString_AS_STRING(result) : 0;
        }
        
        if (!result)
            return 0;
    
        // SQLGetData filled our buffer completely, minus the NULL (if applicable).  cbData is the total length of the
        // data, not the amount already given to us.  (If we read the whole thing, we'd expect the return to have been
        // SQL_SUCCESS and we wouldn't have made it this far.)
        size_t cbHave = MAX_VAR_STACK - null_size;
        memcpy(pch, ab, cbHave);
        pch += cbHave;
    
        ret = SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), nTargetType, pch, cbData - cbHave + null_size, &cbData);
    
        if (!SQL_SUCCEEDED(ret))
        {
            Py_DECREF(result);
            return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
        }
    
        return result;
    }

    // The data didn't fit in the buffer and we have no idea how long it is.  We'll have to keep resizing our buffer
    // and reading until we get it all.
    
    Py_ssize_t elements = min((MAX_VAR_STACK / element_size) * 2, (int)pinfo->column_size);
    
    if (nTargetType == SQL_C_WCHAR)
    {
        result = PyUnicode_FromUnicode(0, elements);
        pch    = result ? (char*)PyUnicode_AsUnicode(result) : 0;
    }
    else
    {
        result = PyString_FromStringAndSize(0, elements);
        pch    = result ? PyString_AS_STRING(result) : 0;
    }
    
    if (!result)
        return 0;

    Py_ssize_t cbHave = MAX_VAR_STACK - null_size;

    memcpy(pch, ab, cbHave);

    for (;;)
    {
        // How much room (in bytes) do we have in the current buffer?  We add null_size because both string and
        // unicode objects allocate room for the NULL.
        Py_ssize_t cbAvailable = elements * element_size - cbHave + null_size;

        ret = SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), nTargetType, pch + cbHave, cbAvailable, &cbData);

        if (ret == SQL_NO_DATA)
        {
            // According to the ODBC documentation, some drivers that don't return SQL_SUCCEESS with the last data.
            // Instead, they return SQL_SUCCESS_WITH_INFO with each call, then SQL_NO_DATA when called *after* the last
            // buffer.  Since we don't know that they properly set cbData either, we set it ourselves and drop into
            // the SQL_SUCCESS case below.

            ret = SQL_SUCCESS;
            cbData = 0;
        }

        if (ret == SQL_SUCCESS)
        {
            // SQL_SUCCESS is returned with the last buffer, so break out of the fetch loop.
            //
            // If we allocated more memory than we finally needed, shrink the objects.  Note that SQLGetData is
            // required to return a buffer length in cbData when returning SQL_SUCCESS (not SQL_NO_TOTAL).  Also
            // remember that cbAvaiable includes room for the NULL terminator but cbData does not.  (The NULL
            // terminator was copied to the buffer, but not included in the returned count.)

            if (cbData + null_size < cbAvailable)
            {
                elements = (cbHave + cbData) / element_size;

                if (nTargetType == SQL_C_WCHAR)
                {
                    if (PyUnicode_Resize(&result, elements) == -1)
                        return 0;
                }
                else
                {
                    if (_PyString_Resize(&result, elements) == -1)
                        return 0;
                }
            }

            break;
        }

        if (!SQL_SUCCEEDED(ret))
        {
            Py_DECREF(result);
            return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
        }

        if (cbData != SQL_NO_TOTAL)
        {
            // We now have the final length.
            elements = cbData / element_size;
        }
        else
        {
            // We still don't know; allocate more memory and try again.
            elements = min(elements * 2, (SQLINTEGER)pinfo->column_size);
        }

        if (nTargetType == SQL_C_WCHAR)
        {
            if (PyUnicode_Resize(&result, elements) == -1)
                return 0;
            pch = (char*)PyUnicode_AsUnicode(result);
        }
        else
        {
            if (_PyString_Resize(&result, elements) ==  -1)
                return 0;
            pch = PyString_AS_STRING(result);
        }

        // Keep track of how much we have read so far.
        cbHave += cbAvailable - null_size;
    }

    return result;
}

static PyObject*
GetDataBuffer(Cursor* cur, Py_ssize_t iCol)
{
    PyObject* str = GetDataString(cur, iCol);

    if (str == Py_None)
        return str;

    PyObject* buffer = 0;

    if (str)
    {
        buffer = PyBuffer_FromObject(str, 0, PyString_GET_SIZE(str));
        Py_DECREF(str);         // If no buffer, release it.  If buffer, the buffer owns it.
    }

    return buffer;
}

static PyObject*
GetDataDecimal(Cursor* cur, int iCol)
{
    // Until we optimize and preallocate the input buffer and bind the columns, I don't see any reason to use
    // SQL_NUMERIC_STRUCT.  It would require us to perform the conversions, but the ODBC drivers are probably more
    // optimized than ours would be.

    ColumnInfo* pinfo = &cur->colinfos[iCol];
    SQLINTEGER cbFetched = 0;
    char* sz = (char*)_alloca(pinfo->column_size + 3); // + '-', '.', and NULL

    if (sz == 0)
        return PyErr_NoMemory();

    if (!SQL_SUCCEEDED(SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), SQL_C_CHAR, sz, pinfo->column_size + 3, &cbFetched)))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    if (cbFetched == SQL_NULL_DATA)
        Py_RETURN_NONE;

    return PyObject_CallFunction(decimal_type, "s", sz);
}

static PyObject*
GetDataBit(Cursor* cur, int iCol)
{
    SQLCHAR ch;
    SQLINTEGER cbFetched;

    if (!SQL_SUCCEEDED(SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), SQL_C_BIT, &ch, sizeof(ch), &cbFetched)))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    if (cbFetched == SQL_NULL_DATA)
        Py_RETURN_NONE;

    if (ch == SQL_TRUE)
        Py_RETURN_TRUE;

    Py_RETURN_FALSE;
}

static PyObject*
GetDataLong(Cursor* cur, int iCol)
{
    ColumnInfo* pinfo = &cur->colinfos[iCol];

    long value = 0;
    SQLINTEGER cbFetched = 0;

    SQLSMALLINT nCType = pinfo->is_unsigned ? SQL_C_ULONG : SQL_C_LONG;

    if (!SQL_SUCCEEDED(SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), nCType, &value, sizeof(value), &cbFetched)))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    if (cbFetched == SQL_NULL_DATA)
        Py_RETURN_NONE;

    if (pinfo->is_unsigned)
        return PyInt_FromLong(*(unsigned long*)&value);

    return PyInt_FromLong(value);
}

static PyObject*
GetDataLongLong(Cursor* cur, int iCol)
{
    ColumnInfo* pinfo = &cur->colinfos[iCol];

    __int64 value = 0;
    SQLINTEGER cbFetched = 0;

    SQLSMALLINT nCType = pinfo->is_unsigned ? SQL_C_UBIGINT : SQL_C_SBIGINT;

    if (!SQL_SUCCEEDED(SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), nCType, &value, sizeof(value), &cbFetched)))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    if (cbFetched == SQL_NULL_DATA)
        Py_RETURN_NONE;

    if (pinfo->is_unsigned)
        return PyLong_FromLongLong(*(unsigned __int64*)&value);

    return PyLong_FromLongLong(*(__int64*)&value);
}

static PyObject*
GetDataDouble(Cursor* cur, int iCol)
{
    double value;
    SQLINTEGER cbFetched = 0;

    if (!SQL_SUCCEEDED(SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), SQL_C_DOUBLE, &value, sizeof(value), &cbFetched)))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    if (cbFetched == SQL_NULL_DATA)
        Py_RETURN_NONE;

    return PyFloat_FromDouble(value);
}


static PyObject*
GetDataTimestamp(Cursor* cur, int iCol)
{
    TIMESTAMP_STRUCT value;

    SQLINTEGER cbFetched = 0;

    if (!SQL_SUCCEEDED(SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), SQL_C_TYPE_TIMESTAMP, &value, sizeof(value), &cbFetched)))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    if (cbFetched == SQL_NULL_DATA)
        Py_RETURN_NONE;

    switch (cur->colinfos[iCol].sql_type)
    {
    case SQL_TYPE_TIME:
        return PyTime_FromTime(value.hour, value.minute, value.second, 0);

    case SQL_TYPE_DATE:
        return PyDate_FromDate(value.year, value.month, value.day);
    }

    return PyDateTime_FromDateAndTime(value.year, value.month, value.day, value.hour, value.minute, value.second, 0);
}


static PyObject*
GetData(Cursor* cur, Py_ssize_t iCol)
{
    // Returns an object representing the value in the row/field.  If 0 is returned, an exception has already been set.
    //
    // The data is assumed to be the default C type for the column's SQL type.

    ColumnInfo* pinfo = &cur->colinfos[iCol];

    switch (pinfo->sql_type)
    {
    case SQL_WCHAR:
    case SQL_WVARCHAR:
    case SQL_WLONGVARCHAR:
    case SQL_CHAR:
    case SQL_VARCHAR:
    case SQL_LONGVARCHAR:
    case SQL_GUID:
        return GetDataString(cur, iCol);

    case SQL_BINARY:
    case SQL_VARBINARY:
    case SQL_LONGVARBINARY:
        return GetDataBuffer(cur, iCol);

    case SQL_DECIMAL:
    case SQL_NUMERIC:
    {
        if (decimal_type == 0)
            break;

        return GetDataDecimal(cur, iCol);
    }

    case SQL_BIT:
        return GetDataBit(cur, iCol);

    case SQL_TINYINT:
    case SQL_SMALLINT:
    case SQL_INTEGER:
        return GetDataLong(cur, iCol);

    case SQL_BIGINT:
        return GetDataLongLong(cur, iCol);

    case SQL_REAL:
    case SQL_FLOAT:
    case SQL_DOUBLE:
        return GetDataDouble(cur, iCol);


    case SQL_TYPE_DATE:
    case SQL_TYPE_TIME:
    case SQL_TYPE_TIMESTAMP:
        return GetDataTimestamp(cur, iCol);
    }

    return RaiseErrorV("HY106", ProgrammingError, "ODBC SQL type %d is not yet supported.  column-index=%zd  type=%d",
                       (int)pinfo->sql_type, iCol, (int)pinfo->sql_type);
}


static PyObject*
Cursor_fetch(Cursor* cur)
{
    // Internal function to fetch a single row and construct a Row object from it.  Used by all of the fetching
    // functions.
    // 
    // Returns a Row object if successful.  If there are no more rows, zero is returned.  If an error occurs, an
    // exception is set and zero is returned.  (To differentiate between the last two, use PyErr_Occurred.)

    SQLRETURN ret = 0;
    Py_ssize_t field_count, i;
    PyObject** apValues;

    Py_BEGIN_ALLOW_THREADS
    ret = SQLFetch(cur->hstmt);
    Py_END_ALLOW_THREADS

    if (GetConnection(cur)->hdbc == SQL_NULL_HANDLE)
    {
        // The connection was closed by another thread in the ALLOW_THREADS block above.
        return RaiseErrorV(0, ProgrammingError, "The cursor's connection was closed.");
    }

    if (ret == SQL_NO_DATA)
        return 0;

    if (!SQL_SUCCEEDED(ret))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    field_count = PyTuple_GET_SIZE(cur->description);

    apValues = (PyObject**)malloc(sizeof(PyObject*) * field_count);

    if (apValues == 0)
        return PyErr_NoMemory();

    for (i = 0; i < field_count; i++)
    {
        PyObject* value = GetData(cur, i);

        if (!value)
        {
            FreeRowValues(i, apValues);
            return 0;
        }

        apValues[i] = value;
    }

    return (PyObject*)Row_New(cur->map_name_to_index, field_count, apValues);
}


static PyObject*
Cursor_fetchlist(Cursor* cur, Py_ssize_t max)
{
    // max
    //   The maximum number of rows to fetch.  If -1, fetch all rows.
    // 
    // Returns a list of Rows.  If there are no rows, an empty list is returned.

    PyObject* results;
    PyObject* row;

    results = PyList_New(0);
    if (!results)
        return 0;

    while (max == -1 || max > 0)
    {
        row = Cursor_fetch(cur);

        if (!row)
        {
            if (PyErr_Occurred())
            {
                Py_DECREF(results);
                return 0;
            }
            break;
        }

        PyList_Append(results, row);
        Py_DECREF(row);

        if (max != -1)
            max--;
    }

    return results;
}

static PyObject*
Cursor_iter(PyObject* self)
{
    Py_INCREF(self);
    return self;
}


static PyObject*
Cursor_iternext(PyObject* self)
{
    // Implements the iterator protocol for cursors.  Fetches the next row.  Returns zero without setting an exception
    // when there are no rows.

    PyObject* result;

    Cursor* cursor = Cursor_Validate(self, CURSOR_REQUIRE_RESULTS | CURSOR_RAISE_ERROR);

    if (!cursor)
        return 0;

    result = Cursor_fetch(cursor);

    return result;
}

static PyObject*
Cursor_fetchone(PyObject* self, PyObject* args)
{
    PyObject* row;
    Cursor* cursor = Cursor_Validate(self, CURSOR_REQUIRE_RESULTS | CURSOR_RAISE_ERROR);
    if (!cursor)
        return 0;

    row = Cursor_fetch(cursor);

    if (!row)
    {
        if (PyErr_Occurred())
            return 0;
        Py_RETURN_NONE;
    }

    return row;
}

static PyObject*
Cursor_fetchall(PyObject* self, PyObject* args)
{
    PyObject* result;
    Cursor* cursor = Cursor_Validate(self, CURSOR_REQUIRE_RESULTS | CURSOR_RAISE_ERROR);
    if (!cursor)
        return 0;

    result = Cursor_fetchlist(cursor, -1);

    return result;
}

static PyObject*
Cursor_fetchmany(PyObject* self, PyObject* args)
{
    long rows;
    PyObject* result;

    Cursor* cursor = Cursor_Validate(self, CURSOR_REQUIRE_RESULTS | CURSOR_RAISE_ERROR);
    if (!cursor)
        return 0;

    rows = cursor->arraysize;
    if (!PyArg_ParseTuple(args, "|l", &rows))
        return 0;

    result = Cursor_fetchlist(cursor, rows);

    return result;
}

static char tables_doc[] =
    "C.tables(table=None, catalog=None, schema=None, tableType=None) --> self\n"
    "\n"
    "Executes SQLTables and creates a results set of tables defined in the data\n"
    "source.\n"
    "\n"
    "The table, catalog, and schema interpret the '_' and '%' characters as\n"
    "wildcards.  The escape character is driver specific, so use\n"
    "`Connection.searchescape`.\n"
    "\n"
    "Each row fetched has the following columns:\n"
    " 0) table_cat: The catalog name.\n"
    " 1) table_schem: The schema name.\n"
    " 2) table_name: The table name.\n"
    " 3) table_type: One of 'TABLE', 'VIEW', SYSTEM TABLE', 'GLOBAL TEMPORARY'\n"
    "    'LOCAL TEMPORARY', 'ALIAS', 'SYNONYM', or a data source-specific type name.";

char* Cursor_tables_kwnames[] = { "table", "catalog", "schema", "tableType", 0 };

static PyObject*
Cursor_tables(PyObject* self, PyObject* args, PyObject* kwargs)
{
    const char* szCatalog = 0;
    const char* szSchema = 0;
    const char* szTableName = 0;
    const char* szTableType = 0;

    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|ssss", Cursor_tables_kwnames, &szTableName, &szCatalog, &szSchema, &szTableType))
        return 0;

    Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
    
    if (!free_results(cur))
        return 0;
    
    SQLRETURN ret = 0;

    Py_BEGIN_ALLOW_THREADS
    ret = SQLTables(cur->hstmt, (SQLCHAR*)szCatalog, SQL_NTS, (SQLCHAR*)szSchema, SQL_NTS, 
                    (SQLCHAR*)szTableName, SQL_NTS, (SQLCHAR*)szTableType, SQL_NTS);
    Py_END_ALLOW_THREADS

    if (!SQL_SUCCEEDED(ret))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    SQLSMALLINT cCols;
    if (!SQL_SUCCEEDED(SQLNumResultCols(cur->hstmt, &cCols)))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
    
    if (!PrepareResults(cur, cCols))
        return 0;

    if (!create_name_map(cur, cCols))
        return 0;

    // Return the cursor so the results can be iterated over directly.
    Py_INCREF(cur);
    return (PyObject*)cur;
}


static char columns_doc[] =
    "C.columns(table=None, catalog=None, schema=None, column=None)\n\n"
    "Creates a results set of column names in specified tables by executing the ODBC SQLColumns function.\n"
    "Each row fetched has the following columns:\n"
    "  0) table_cat\n"
    "  1) table_schem\n"
    "  2) table_name\n"
    "  3) column_name\n"
    "  4) data_type\n"
    "  5) type_name\n"
    "  6) column_size\n"
    "  7) buffer_length\n"
    "  8) decimal_digits\n"
    "  9) num_prec_radix\n"
    " 10) nullable\n"
    " 11) remarks\n"
    " 12) column_def\n"
    " 13) sql_data_type\n"
    " 14) sql_datetime_sub\n"
    " 15) char_octet_length\n"
    " 16) ordinal_position\n"
    " 17) is_nullable";

char* Cursor_column_kwnames[] = { "table", "catalog", "schema", "column", 0 };

static PyObject*
Cursor_columns(PyObject* self, PyObject* args, PyObject* kwargs)
{
    const char* szCatalog = 0;
    const char* szSchema  = 0;
    const char* szTable   = 0;
    const char* szColumn  = 0;

    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|ssss", Cursor_column_kwnames, &szTable, &szCatalog, &szSchema, &szColumn))
        return 0;

    Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
    
    if (!free_results(cur))
        return 0;
    
    SQLRETURN ret = 0;

    Py_BEGIN_ALLOW_THREADS
    ret = SQLColumns(cur->hstmt, (SQLCHAR*)szCatalog, SQL_NTS, (SQLCHAR*)szSchema, SQL_NTS, (SQLCHAR*)szTable, SQL_NTS, (SQLCHAR*)szColumn, SQL_NTS);
    Py_END_ALLOW_THREADS

    if (!SQL_SUCCEEDED(ret))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    SQLSMALLINT cCols;
    if (!SQL_SUCCEEDED(SQLNumResultCols(cur->hstmt, &cCols)))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
    
    if (!PrepareResults(cur, cCols))
        return 0;

    if (!create_name_map(cur, cCols))
        return 0;

    // Return the cursor so the results can be iterated over directly.
    Py_INCREF(cur);
    return (PyObject*)cur;
}


static char statistics_doc[] =
    "C.statistics(catalog=None, schema=None, unique=False, quick=True) --> self\n\n"
    "Creates a results set of statistics about a single table and the indexes associated with \n"
    "the table by executing SQLStatistics.\n"
    "unique\n"
    "  If True, only unique indexes are retured.  Otherwise all indexes are returned.\n"
    "quick\n"
    "  If True, CARDINALITY and PAGES are returned  only if they are readily available\n"
    "  from the server\n"
    "\n"
    "Each row fetched has the following columns:\n\n"
    "  0) table_cat\n"
    "  1) table_schem\n"
    "  2) table_name\n"
    "  3) non_unique\n"
    "  4) index_qualifier\n"
    "  5) index_name\n"
    "  6) type\n"
    "  7) ordinal_position\n"
    "  8) column_name\n"
    "  9) asc_or_desc\n"
    " 10) cardinality\n"
    " 11) pages\n"
    " 12) filter_condition";

char* Cursor_statistics_kwnames[] = { "table", "catalog", "schema", "unique", "quick", 0 };

static PyObject*
Cursor_statistics(PyObject* self, PyObject* args, PyObject* kwargs)
{
    const char* szCatalog = 0;
    const char* szSchema  = 0;
    const char* szTable   = 0;
    PyObject* pUnique = Py_False;
    PyObject* pQuick  = Py_True;

    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|ssOO", Cursor_statistics_kwnames, &szTable, &szCatalog, &szSchema,
                                     &pUnique, &pQuick))
        return 0;

    Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
    
    if (!free_results(cur))
        return 0;
    
    SQLUSMALLINT nUnique   = PyObject_IsTrue(pUnique) ? SQL_INDEX_UNIQUE : SQL_INDEX_ALL;
    SQLUSMALLINT nReserved = PyObject_IsTrue(pQuick)  ? SQL_QUICK : SQL_ENSURE;

    SQLRETURN ret = 0;

    Py_BEGIN_ALLOW_THREADS
    ret = SQLStatistics(cur->hstmt, (SQLCHAR*)szCatalog, SQL_NTS, (SQLCHAR*)szSchema, SQL_NTS, (SQLCHAR*)szTable, SQL_NTS, 
                        nUnique, nReserved);
    Py_END_ALLOW_THREADS

    if (!SQL_SUCCEEDED(ret))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    SQLSMALLINT cCols;
    if (!SQL_SUCCEEDED(SQLNumResultCols(cur->hstmt, &cCols)))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
    
    if (!PrepareResults(cur, cCols))
        return 0;

    if (!create_name_map(cur, cCols))
        return 0;

    // Return the cursor so the results can be iterated over directly.
    Py_INCREF(cur);
    return (PyObject*)cur;
}


static char rowIdColumns_doc[] =
    "C.rowIdColumns(table, catalog=None, schema=None, nullable=True) -->\n\n"
    "Executes SQLSpecialColumns with SQL_BEST_ROWID which creates a result set of columns that\n"
    "uniquely identify a row\n\n"
    "Each row fetched has the following columns:\n"
    " 0) scope\n"
    " 1) column_name\n"
    " 2) data_type\n"
    " 3) type_name\n"
    " 4) column_size\n"
    " 5) buffer_length\n"
    " 6) decimal_digits\n"
    " 7) pseudo_column";

static char rowVerColumns_doc[] =
    "C.rowIdColumns(table, catalog=None, schema=None, nullable=True) --> self\n\n"
    "Executes SQLSpecialColumns with SQL_ROWVER which creates a result set of columns that\n"
    "are automatically updated when any value in the row is updated.\n\n"
    "Each row fetched has the following columns:\n"
    " 0) scope\n"
    " 1) column_name\n"
    " 2) data_type\n"
    " 3) type_name\n"
    " 4) column_size\n"
    " 5) buffer_length\n"
    " 6) decimal_digits\n"
    " 7) pseudo_column";

char* Cursor_specialColumn_kwnames[] = { "table", "catalog", "schema", "nullable", 0 };

static PyObject*
_specialColumns(PyObject* self, PyObject* args, PyObject* kwargs, int nIdType)
{
    const char* szTable;
    const char* szCatalog = 0;
    const char* szSchema  = 0;
    PyObject* pNullable = Py_True;

    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|ssO", Cursor_specialColumn_kwnames, &szTable, &szCatalog, &szSchema, &pNullable))
        return 0;

    Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
    
    if (!free_results(cur))
        return 0;
    
    SQLRETURN ret = 0;

    SQLSMALLINT nNullable = PyObject_IsTrue(pNullable) ? SQL_NULLABLE : SQL_NO_NULLS;

    Py_BEGIN_ALLOW_THREADS
    ret = SQLSpecialColumns(cur->hstmt, nIdType, (SQLCHAR*)szCatalog, SQL_NTS, (SQLCHAR*)szSchema, SQL_NTS, (SQLCHAR*)szTable, SQL_NTS,
                            SQL_SCOPE_TRANSACTION, nNullable);
    Py_END_ALLOW_THREADS

    if (!SQL_SUCCEEDED(ret))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    SQLSMALLINT cCols;
    if (!SQL_SUCCEEDED(SQLNumResultCols(cur->hstmt, &cCols)))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
    
    if (!PrepareResults(cur, cCols))
        return 0;

    if (!create_name_map(cur, cCols))
        return 0;

    // Return the cursor so the results can be iterated over directly.
    Py_INCREF(cur);
    return (PyObject*)cur;
}

static PyObject*
Cursor_rowIdColumns(PyObject* self, PyObject* args, PyObject* kwargs)
{
    return _specialColumns(self, args, kwargs, SQL_BEST_ROWID);
}

static PyObject*
Cursor_rowVerColumns(PyObject* self, PyObject* args, PyObject* kwargs)
{
    return _specialColumns(self, args, kwargs, SQL_ROWVER);
}


static char primaryKeys_doc[] =
    "C.primaryKeys(table, catalog=None, schema=None) --> self\n\n"
    "Creates a results set of column names that make up the primary key for a table\n"
    "by executing the SQLPrimaryKeys function.\n"
    "Each row fetched has the following columns:\n"
    " 0) table_cat\n"
    " 1) table_schem\n"
    " 2) table_name\n"
    " 3) column_name\n"
    " 4) key_seq\n"
    " 5) pk_name";

char* Cursor_primaryKeys_kwnames[] = { "table", "catalog", "schema", 0 };

static PyObject*
Cursor_primaryKeys(PyObject* self, PyObject* args, PyObject* kwargs)
{
    const char* szTable;
    const char* szCatalog = 0;
    const char* szSchema  = 0;

    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|ss", Cursor_primaryKeys_kwnames, &szTable, &szCatalog, &szSchema))
        return 0;

    Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
    
    if (!free_results(cur))
        return 0;
    
    SQLRETURN ret = 0;

    Py_BEGIN_ALLOW_THREADS
    ret = SQLPrimaryKeys(cur->hstmt, (SQLCHAR*)szCatalog, SQL_NTS, (SQLCHAR*)szSchema, SQL_NTS, (SQLCHAR*)szTable, SQL_NTS);
    Py_END_ALLOW_THREADS

    if (!SQL_SUCCEEDED(ret))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    SQLSMALLINT cCols;
    if (!SQL_SUCCEEDED(SQLNumResultCols(cur->hstmt, &cCols)))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
    
    if (!PrepareResults(cur, cCols))
        return 0;

    if (!create_name_map(cur, cCols))
        return 0;

    // Return the cursor so the results can be iterated over directly.
    Py_INCREF(cur);
    return (PyObject*)cur;
}


static char foreignKeys_doc[] =
    "C.foreignKeys(table=None, catalog=None, schema=None,\n"
    "            foreignTable=None, foreignCatalog=None, foreignSchema=None) --> self\n\n"
    "Executes the SQLForeignKeys function and creates a results set of column names\n"
    "that are foreign keys in the specified table (columns in the specified table\n"
    "that refer to primary keys in other tables) or foreign keys in other tables\n"
    "that refer to the primary key in the specified table.\n\n"
    "Each row fetched has the following columns:\n"
    "  0) pktable_cat\n"
    "  1) pktable_schem\n"
    "  2) pktable_name\n"
    "  3) pkcolumn_name\n"
    "  4) fktable_cat\n"
    "  5) fktable_schem\n"
    "  6) fktable_name\n"
    "  7) fkcolumn_name\n"
    "  8) key_seq\n"
    "  9) update_rule\n"
    " 10) delete_rule\n"
    " 11) fk_name\n"
    " 12) pk_name\n"
    " 13) deferrability";

char* Cursor_foreignKeys_kwnames[] = { "table", "catalog", "schema", "foreignTable", "foreignCatalog", "foreignSchema", 0 };

static PyObject*
Cursor_foreignKeys(PyObject* self, PyObject* args, PyObject* kwargs)
{
    const char* szTable          = 0;
    const char* szCatalog        = 0;
    const char* szSchema         = 0;
    const char* szForeignTable   = 0;
    const char* szForeignCatalog = 0;
    const char* szForeignSchema  = 0;

    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|ssssss", Cursor_foreignKeys_kwnames, &szTable, &szCatalog, &szSchema,
        &szForeignTable, &szForeignCatalog, &szForeignSchema))
        return 0;

    Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
    
    if (!free_results(cur))
        return 0;
    
    SQLRETURN ret = 0;

    Py_BEGIN_ALLOW_THREADS
    ret = SQLForeignKeys(cur->hstmt, (SQLCHAR*)szCatalog, SQL_NTS, (SQLCHAR*)szSchema, SQL_NTS, (SQLCHAR*)szTable, SQL_NTS,
                         (SQLCHAR*)szForeignCatalog, SQL_NTS, (SQLCHAR*)szForeignSchema, SQL_NTS, (SQLCHAR*)szForeignTable, SQL_NTS);
    Py_END_ALLOW_THREADS

    if (!SQL_SUCCEEDED(ret))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    SQLSMALLINT cCols;
    if (!SQL_SUCCEEDED(SQLNumResultCols(cur->hstmt, &cCols)))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
    
    if (!PrepareResults(cur, cCols))
        return 0;

    if (!create_name_map(cur, cCols))
        return 0;

    // Return the cursor so the results can be iterated over directly.
    Py_INCREF(cur);
    return (PyObject*)cur;
}

static char getTypeInfo_doc[] =
    "C.getTypeInfo(sqlType=None) --> self\n\n"
    "Executes SQLGetTypeInfo a creates a result set with information about the\n"
    "specified data type or all data types supported by the ODBC driver if not\n"
    "specified.\n\n"
    "Each row fetched has the following columns:\n"
    " 0) type_name\n"
    " 1) data_type\n"
    " 2) column_size\n"
    " 3) literal_prefix\n"
    " 4) literal_suffix\n"
    " 5) create_params\n"
    " 6) nullable\n"
    " 7) case_sensitive\n"
    " 8) searchable\n"
    " 9) unsigned_attribute\n"
    "10) fixed_prec_scale\n"
    "11) auto_unique_value\n"
    "12) local_type_name\n"
    "13) minimum_scale\n"
    "14) maximum_scale\n"
    "15) sql_data_type\n"
    "16) sql_datetime_sub\n"
    "17) num_prec_radix\n"
    "18) interval_precision";

static PyObject*
Cursor_getTypeInfo(PyObject* self, PyObject* args, PyObject* kwargs)
{
    int nDataType = SQL_ALL_TYPES;

    if (!PyArg_ParseTuple(args, "|i", &nDataType))
        return 0;

    Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
    
    if (!free_results(cur))
        return 0;
    
    SQLRETURN ret = 0;

    Py_BEGIN_ALLOW_THREADS
    ret = SQLGetTypeInfo(cur->hstmt, nDataType);
    Py_END_ALLOW_THREADS

    if (!SQL_SUCCEEDED(ret))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    SQLSMALLINT cCols;
    if (!SQL_SUCCEEDED(SQLNumResultCols(cur->hstmt, &cCols)))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
    
    if (!PrepareResults(cur, cCols))
        return 0;

    if (!create_name_map(cur, cCols))
        return 0;

    // Return the cursor so the results can be iterated over directly.
    Py_INCREF(cur);
    return (PyObject*)cur;
}

/*
static PyObject*
Cursor_nextset(PyObject* self, PyObject* args)
{
    PyErr_SetString(NotSupportedError, "nextset() is not supported");
    return 0;
}
*/

static char procedures_doc[] =
    "C.procedures(procedure=None, catalog=None, schema=None) --> self\n\n"
    "Executes SQLProcedures and creates a result set of information about the\n"
    "procedures in the data source.\n"
    "Each row fetched has the following columns:\n"
    " 0) procedure_cat\n"
    " 1) procedure_schem\n"
    " 2) procedure_name\n"
    " 3) num_input_params\n"
    " 4) num_output_params\n"
    " 5) num_result_sets\n"
    " 6) remarks\n"
    " 7) procedure_type";

char* Cursor_procedures_kwnames[] = { "procedure", "catalog", "schema", 0 };

static PyObject*
Cursor_procedures(PyObject* self, PyObject* args, PyObject* kwargs)
{
    const char* szProcedure = 0;
    const char* szCatalog   = 0;
    const char* szSchema    = 0;

    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|sss", Cursor_procedures_kwnames, &szProcedure, &szCatalog, &szSchema))
        return 0;

    Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
    
    if (!free_results(cur))
        return 0;
    
    SQLRETURN ret = 0;

    Py_BEGIN_ALLOW_THREADS
    ret = SQLProcedures(cur->hstmt, (SQLCHAR*)szCatalog, SQL_NTS, (SQLCHAR*)szSchema, SQL_NTS, (SQLCHAR*)szProcedure, SQL_NTS);
    Py_END_ALLOW_THREADS

    if (!SQL_SUCCEEDED(ret))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);

    SQLSMALLINT cCols;
    if (!SQL_SUCCEEDED(SQLNumResultCols(cur->hstmt, &cCols)))
        return RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
    
    if (!PrepareResults(cur, cCols))
        return 0;

    if (!create_name_map(cur, cCols))
        return 0;

    // Return the cursor so the results can be iterated over directly.
    Py_INCREF(cur);
    return (PyObject*)cur;
}

static PyObject*
Cursor_ignored(PyObject* self, PyObject* args)
{
    Py_RETURN_NONE;
}

static char rowcount_doc[] =
    "This read-only attribute specifies the number of rows the last DML statement\n"
    " (INSERT, UPDATE, DELETE) affected.  This is set to -1 for SELECT statements.";

static char description_doc[] =
    "This read-only attribute is a sequence of 7-item sequences.  Each of these\n" \
    "sequences contains information describing one result column: (name, type_code,\n" \
    "display_size, internal_size, precision, scale, null_ok).  All values except\n" \
    "name, type_code, and internal_size are None.  The type_code entry will be the\n" \
    "type object used to create values for that column (e.g. `str` or\n" \
    "`datetime.datetime`).\n" \
    "\n" \
    "This attribute will be None for operations that do not return rows or if the\n" \
    "cursor has not had an operation invoked via the execute() method yet.\n" \
    "\n" \
    "The type_code can be interpreted by comparing it to the Type Objects defined in\n" \
    "the DB API and defined the pyodbc module: Date, Time, Timestamp, Binary,\n" \
    "STRING, BINARY, NUMBER, and DATETIME.";

static char arraysize_doc[] =
    "This read/write attribute specifies the number of rows to fetch at a time with\n" \
    "fetchmany(). It defaults to 1 meaning to fetch a single row at a time.";

static char connection_doc[] =
    "This read-only attribute return a reference to the Connection object on which\n" \
    "the cursor was created.\n" \
    "\n" \
    "The attribute simplifies writing polymorph code in multi-connection\n" \
    "environments.";

static PyMemberDef Cursor_members[] =
{
    {"rowcount",    T_INT,       offsetof(Cursor, rowcount),        READONLY, rowcount_doc },
    {"description", T_OBJECT_EX, offsetof(Cursor, description),     READONLY, description_doc },
    {"arraysize",   T_INT,       offsetof(Cursor, arraysize),       0,        arraysize_doc },
    {"connection",  T_OBJECT_EX, offsetof(Cursor, cnxn),            READONLY, connection_doc },
    { 0 }
};


static char executemany_doc[] =
    "executemany(sql, seq_of_params) --> Cursor | count | None\n" \
    "\n" \
    "Prepare a database query or command and then execute it against all parameter\n" \
    "sequences  found in the sequence seq_of_params.\n" \
    "\n" \
    "Only the result of the final execution is returned.  See `execute` for a\n" \
    "description of parameter passing the return value.";
    
/*
static char nextset_doc[] = "Raises a NotSupportedError error.";
*/

static char ignored_doc[] = "Ignored.";

static char fetchone_doc[] =
    "fetchone() --> Row | None\n" \
    "\n" \
    "Fetch the next row of a query result set, returning a single Row instance, or\n" \
    "None when no more data is available.\n" \
    "\n" \
    "A ProgrammingError exception is raised if the previous call to execute() did\n" \
    "not produce any result set or no call was issued yet.";

static char fetchall_doc[] =
    "fetchmany(size=cursor.arraysize) --> list of Rows\n" \
    "\n" \
    "Fetch the next set of rows of a query result, returning a list of Row\n" \
    "instances. An empty list is returned when no more rows are available.\n" \
    "\n" \
    "The number of rows to fetch per call is specified by the parameter.  If it is\n" \
    "not given, the cursor's arraysize determines the number of rows to be\n" \
    "fetched. The method should try to fetch as many rows as indicated by the size\n" \
    "parameter. If this is not possible due to the specified number of rows not\n" \
    "being available, fewer rows may be returned.\n" \
    "\n" \
    "A ProgrammingError exception is raised if the previous call to execute() did\n" \
    "not produce any result set or no call was issued yet.";

static char fetchmany_doc[] =
    "fetchmany() --> list of Rows\n" \
    "\n" \
    "Fetch all remaining rows of a query result, returning them as a list of Rows.\n" \
    "An empty list is returned if there are no more rows.\n" \
    "\n" \
    "A ProgrammingError exception is raised if the previous call to execute() did\n" \
    "not produce any result set or no call was issued yet.";

static PyMethodDef Cursor_methods[] =
{
    { "close",         (PyCFunction)Cursor_close,         METH_NOARGS,                close_doc         },
    { "execute",       (PyCFunction)Cursor_execute,       METH_VARARGS,               execute_doc       },
    { "executemany",   (PyCFunction)Cursor_executemany,   METH_VARARGS,               executemany_doc   },
    { "setinputsizes", (PyCFunction)Cursor_ignored,       METH_VARARGS,               ignored_doc       },
    { "setoutputsize", (PyCFunction)Cursor_ignored,       METH_VARARGS,               ignored_doc       },
    { "fetchone",      (PyCFunction)Cursor_fetchone,      METH_NOARGS,                fetchone_doc      },
    { "fetchall",      (PyCFunction)Cursor_fetchall,      METH_NOARGS,                fetchall_doc      },
    { "fetchmany",     (PyCFunction)Cursor_fetchmany,     METH_VARARGS,               fetchmany_doc     },
    { "tables",        (PyCFunction)Cursor_tables,        METH_VARARGS|METH_KEYWORDS, tables_doc        },
    { "columns",       (PyCFunction)Cursor_columns,       METH_VARARGS|METH_KEYWORDS, columns_doc       },
    { "statistics",    (PyCFunction)Cursor_statistics,    METH_VARARGS|METH_KEYWORDS, statistics_doc    },
    { "rowIdColumns",  (PyCFunction)Cursor_rowIdColumns,  METH_VARARGS|METH_KEYWORDS, rowIdColumns_doc  },
    { "rowVerColumns", (PyCFunction)Cursor_rowVerColumns, METH_VARARGS|METH_KEYWORDS, rowVerColumns_doc },
    { "primaryKeys",   (PyCFunction)Cursor_primaryKeys,   METH_VARARGS|METH_KEYWORDS, primaryKeys_doc   },
    { "foreignKeys",   (PyCFunction)Cursor_foreignKeys,   METH_VARARGS|METH_KEYWORDS, foreignKeys_doc   },
    { "getTypeInfo",   (PyCFunction)Cursor_getTypeInfo,   METH_VARARGS|METH_KEYWORDS, getTypeInfo_doc   },
    { "procedures",    (PyCFunction)Cursor_procedures,    METH_VARARGS|METH_KEYWORDS, procedures_doc    },
    { 0, 0, 0, 0 }
};

static char cursor_doc[] =
    "Cursor objects represent a database cursor, which is used to manage the context\n" \
    "of a fetch operation.  Cursors created from the same connection are not\n" \
    "isolated, i.e., any changes done to the database by a cursor are immediately\n" \
    "visible by the other cursors.  Cursors created from different connections are\n" \
    "isolated.\n" \
    "\n" \
    "Cursors implement the iterator protocol, so results can be iterated:\n" \
    "\n" \
    "  cursor.execute(sql)\n" \
    "  for row in cursor:\n" \
    "     print row[0]";
    
PyTypeObject CursorType =
{
    PyObject_HEAD_INIT(0)
    0,                                                      // ob_size
    "pyodbc.Cursor",                                        // tp_name
    sizeof(Cursor),                                         // tp_basicsize
    0,                                                      // tp_itemsize
    (destructor)Cursor_dealloc,                             // destructor tp_dealloc
    0,                                                      // tp_print
    0,                                                      // tp_getattr
    0,                                                      // tp_setattr
    0,                                                      // tp_compare
    0,                                                      // tp_repr
    0,                                                      // tp_as_number
    0,                                                      // tp_as_sequence
    0,                                                      // tp_as_mapping
    0,                                                      // tp_hash
    0,                                                      // tp_call
    0,                                                      // tp_str
    0,                                                      // tp_getattro
    0,                                                      // tp_setattro
    0,                                                      // tp_as_buffer
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_ITER,              // tp_flags
    cursor_doc,                                             // tp_doc
    0,                                                      // tp_traverse
    0,                                                      // tp_clear
    0,                                                      // tp_richcompare
    0,                                                      // tp_weaklistoffset
    (getiterfunc)Cursor_iter,                               // tp_iter
    (iternextfunc)Cursor_iternext,                          // tp_iternext
    Cursor_methods,                                         // tp_methods
    Cursor_members,                                         // tp_members
    0,                                                      // tp_getset
    0,                                                      // tp_base
    0,                                                      // tp_dict
    0,                                                      // tp_descr_get
    0,                                                      // tp_descr_set
    0,                                                      // tp_dictoffset
    0,                                                      // tp_init
    0,                                                      // tp_alloc
    0,                                                      // tp_new
    0,                                                      // tp_free
    0,                                                      // tp_is_gc
    0,                                                      // tp_bases
    0,                                                      // tp_mro
    0,                                                      // tp_cache
    0,                                                      // tp_subclasses
    0,                                                      // tp_weaklist
};

Cursor*
Cursor_New(Connection* cnxn)
{
    // Exported to allow the connection class to create cursors.

    Cursor* cur = PyObject_NEW(Cursor, &CursorType);

    if (cur)
    {
        cur->cnxn              = (PyObject*)cnxn;
        cur->hstmt             = SQL_NULL_HANDLE;
        cur->description       = Py_None;
        cur->colinfos            = 0;
        cur->arraysize         = 1;
        cur->rowcount          = -1;
        cur->map_name_to_index = 0;

        Py_INCREF(cnxn);
        Py_INCREF(cur->description);

        if (!SQL_SUCCEEDED(SQLAllocHandle(SQL_HANDLE_STMT, cnxn->hdbc, &cur->hstmt)))
        {
            RaiseErrorFromHandle(cnxn->hdbc, SQL_NULL_HANDLE);
            Py_DECREF(cur);
            return 0;
        }
    }

    return cur;
}

void
Cursor_init()
{
    PyDateTime_IMPORT;    
}

static unsigned int toint(PyObject* pDigits, int iStart, int iStop)
{
    UINT value = 0;
    for (int i = iStart; i < iStop; i++)
    {
        value *= 10;
        value += (UINT)PyInt_AsLong(PyTuple_GET_ITEM(pDigits, i));
    }
    return value;
}

static bool multiply(unsigned int* pn, int count, UINT multiplier)
{
    UINT64 carry = 0;

    for (int i = 0; i < count; i++)
    {
        UINT64 product = pn[i] * (UINT64)multiplier + carry;
        carry = product >> 32;
        pn[i] = (UINT)product;
    }

    // If we still have some carry, we have overflowed.
    return carry == 0;
}

static bool add(UINT* pn, int count, UINT addend)
{
    UINT64 sum = pn[0] + addend;
    UINT64 carry = sum >> 32; // (make int64 to widen the addition in the loop)

    pn[0] = (UINT)sum;

    for (int i = 1; i < count && carry != 0; i++)
    {
        sum = pn[i] + carry;
        pn[i] = (UINT)sum;
        carry = sum >> 32;
    }

    // If we still have some carry, we have overflowed.
    return carry == 0;
}

// We represent the ns.val characters as ints.  This is the number of ints needed.
//C_ASSERT((SQL_MAX_NUMERIC_LEN % sizeof(int)) == 0);
static const int MAX_INTS = (SQL_MAX_NUMERIC_LEN / sizeof(int));

bool decimal_to_struct(PyObject* pTuple, SQL_NUMERIC_STRUCT& ns)
{
    // Fills out the SQL_NUMERIC_STRUCT from a decimal.Decimal tuple.
    //
    // This version is very IA32 oriented, including the byte ordering.  We need some #error lines to check for
    // ordering and sizes.

    memset(&ns, 0, sizeof(SQL_NUMERIC_STRUCT));

    // We deal with the value as integers to speed up the math.  Note that this *requires* us to be on a little-endian
    // machine.
    UINT* pn = (UINT*)ns.val;

    PyObject* pSign   = PyTuple_GET_ITEM(pTuple, 0);
    PyObject* pDigits = PyTuple_GET_ITEM(pTuple, 1);
    PyObject* pScale  = PyTuple_GET_ITEM(pTuple, 2);

    ns.precision = PyTuple_GET_SIZE(pDigits); 
    ns.sign      = !PyInt_AsLong(pSign);
    ns.scale     = (SQLSCHAR)-PyInt_AsLong(pScale);

    const int cDigits = ns.precision;

    int iStart = 0;
    int iStop  = cDigits % 9;
    if (iStop == 0)
        iStop = min(cDigits, 9);

    pn[0] = toint(pDigits, iStart, iStop);

    iStart = iStop;
    while (iStart < cDigits)
    {
        int n = toint(pDigits, iStart, iStart + 9);

        if (!multiply(pn, MAX_INTS, 1000000000) || !add(pn, MAX_INTS, n))
        {
            RaiseErrorV(0, 0, "Decimal precision too large.");
            return false;
        }
        
        iStart += 9;
    }

    return true;
}
